#include <stdio.h>

#include "zchaff/SAT_C.h"

int sol[9*9];

/*
 * add clauses describing val(i) != v
 */
void add_negative(SAT_Manager m, int i, int v)
{
  int c[4];
  i *= 4;
  i++;
  /*
   * X !=i means:
   *    X0 != i0 OR X1 != i1 ... OR X3 != i3
   * Xn != in translates to Xn if in==0 and to !Xn if in==1
   * So we have a simple clause here.
   * In the code below:
   * i*2 means Xn
   * i*2+1 means !Xn
   * We have done i*=4 because variable i has boolean variables i*4 ... i*4+3
   * and i++ because zchaff starts to count from 1 not 0
   */
  c[0] = (v&1)? i*2+1 : i*2; i++;
  c[1] = (v&2)? i*2+1 : i*2; i++;
  c[2] = (v&4)? i*2+1 : i*2; i++;
  c[3] = (v&8)? i*2+1 : i*2;
  SAT_AddClause(m, c, 4, 0);
}

/*
 * add clauses describing val(i) == v
 */
void add_positive(SAT_Manager m, int i, int v)
{
  int c;
  i *= 4;
  i++;
  /* X = i means:
   * X0 = i0 AND X1 = i1 ... AND X3 = i3
   * Xn = in translates to Xn if in==1 and to !Xn if in==0
   * So we have 4 unity clauses here.
   * In the code below:
   * i*2 means Xn
   * i*2+1 means !Xn
   * We have done i*=4 because variable i has boolean variables i*4 ... i*4+3
   * and i++ because zchaff starts to count from 1 not 0
   */
  c = (v&1)? i*2 : i*2+1; i++; SAT_AddClause(m, &c, 1, 0);
  c = (v&2)? i*2 : i*2+1; i++; SAT_AddClause(m, &c, 1, 0);
  c = (v&4)? i*2 : i*2+1; i++; SAT_AddClause(m, &c, 1, 0);
  c = (v&8)? i*2 : i*2+1; SAT_AddClause(m, &c, 1, 0);
}

/*
 * add clauses describing val(i) != val(j)
 */
void add_var_diff(SAT_Manager m, int i, int j)
{
  int c[8];
  int l;
  i *= 4;
  j *= 4;

  i++;
  j++;

  /* we have 2 vars: X and Y
   * that have 4 bits, so we have X1, X2, X3, X4 and Y1, Y2, Y3, Y4
   * X != Y means:
   *    X1 != Y1 OR X2 != Y2 OR X3 != Y3 OR X4 != Y4
   * X1 != Y1 means (X1 AND !Y1) OR (!X1 AND Y1) is true
   * We want a CNF form for that, so we apply the wonderful rule:
   *    (A AND B) OR C = (A OR C) AND (B OR C)
   * which gives an exponential blow up, but we don't care here.
   * We finally get by multiple applications of the rule that:
   *   X1 != Y1 means (X1 OR Y1) AND (!X1 OR !Y1)
   * We have the same for X2 != Y2 etc.
   * Now we have:
   *   ((X1 OR Y1) AND (!X1 OR !Y1)) OR ((X2 OR Y2) AND (!X2 OR !Y2)) OR
   *   ((X3 OR Y3) AND (!X3 OR !Y3)) OR ((X4 OR Y4) AND (!X4 OR !Y4))
   * We still have to apply the same rule to take out the AND and take
   * in the OR.
   * We finally find 16 clauses, of shape:
   *   T1i(X1, Y1) OR T2i(X2, Y2) OR T4i(X3, Y3) OR T8i(X4, Y4)
   * where Tji is true if bit j of integer i is 1
   * and when Tji is true then Tji(Xn, Yn) = Xn OR Yn
   * and when Tji is false then Tji(Xn, Yn) = !Xn OR !Yn
   * This loop does create the 16 clauses by using this Tji stuff.
   * variable ii to ii+3 represent the 4 bits of X
   * variable jj to jj+3 represent the 4 bits of Y
   * ii*2 means Xn
   * ii*2+1 means !Xn
   * We have done i*=4 because variable i has boolean variables i*4 ... i*4+3
   * and i++ because zchaff starts to count from 1 not 0
   * (in this code we first put the Xn boolean variables followed
   * by the Yn ones)
   */
  for (l=0; l<16; l++) {
    int ii = i;
    int jj = j;
    int *cc = c;

    *cc++ = (l&1) ? ii*2 : ii*2+1; ii++;
    *cc++ = (l&2) ? ii*2 : ii*2+1; ii++;
    *cc++ = (l&4) ? ii*2 : ii*2+1; ii++;
    *cc++ = (l&8) ? ii*2 : ii*2+1;

    *cc++ = (l&1) ? jj*2 : jj*2+1; jj++;
    *cc++ = (l&2) ? jj*2 : jj*2+1; jj++;
    *cc++ = (l&4) ? jj*2 : jj*2+1; jj++;
    *cc++ = (l&8) ? jj*2 : jj*2+1;

    SAT_AddClause(m, c, 8, 0);
  }
}

void force_value(SAT_Manager m, int i)
{
  /* we handle 9x9 sudoku, so a variable can take values 1..9
   * which means it cannot take values 0 and 10..15
   * (we code on 4 bits, so we may have values 0..15)
   * which translates as:
   * X != 0 AND X != 10 AND ... AND X != 15
   * Each inequality has then to be put as a CNF clause.
   */
  add_negative(m, i, 15);
  add_negative(m, i, 14);
  add_negative(m, i, 13);
  add_negative(m, i, 12);
  add_negative(m, i, 11);
  add_negative(m, i, 10);
  add_negative(m, i, 0);
}

/* return 0 if no solution, 1 if solution */
int sat_solve(SAT_Manager m)
{
  int i;

  i = SAT_Solve(m);
  if (i == SATISFIABLE) {
    for (i=0; i<9*9; i++) {
      int res = 0;
      /* to get the value of variable i is simply to put the 4 corresponding
       * bits together
       * The code below does exactly that.
       * For i, its boolean variables are i*4+1 ... i*4 + 4
       */
      if(SAT_GetVarAsgnment(m, i*4+1) == 1) res |= 1;
      if(SAT_GetVarAsgnment(m, i*4+2) == 1) res |= 2;
      if(SAT_GetVarAsgnment(m, i*4+3) == 1) res |= 4;
      if(SAT_GetVarAsgnment(m, i*4+4) == 1) res |= 8;
      printf("%d%c", res, i % 9 == 8 ? '\n' : ' ');
      sol[i] = res;
    }
    i = SATISFIABLE;
  } else if (i == UNSATISFIABLE) {
    printf("no solution\n");
  } else {
    printf("error, %d\n", i);
  }

  return i == SATISFIABLE;
}

int cube_to_var(int cube, int v)
{
  /* variables in a cube are numbered as follow:
   * 0 1 2
   * 3 4 5
   * 6 7 8
   * little cubes in the big cube is also numbered this way.
   * This leads to the following code. We first get the column
   * and line of the little cube, then deduce its topleft
   * variable, and then we find the x and y in the little cube
   * of v and we finally get the variable we seek based on that.
   */
  int x_cube = cube % 3;
  int y_cube = cube / 3;
  int top_left = y_cube * 9 * 3 + x_cube * 3;
  int x_v = v % 3;
  int y_v = v / 3;
  return top_left + y_v * 9 + x_v;
}

/*
 * this function forces variables i and j in the cube 'cube'
 * to be different
 */
void do_cube(SAT_Manager m, int cube, int i, int j)
{
  int v1 = cube_to_var(cube, i);
  int v2 = cube_to_var(cube, j);
  add_var_diff(m, v1, v2);
}

void forbid_sol(SAT_Manager m)
{
  int c[9*9*4];
  int *cc = c;
  int i;

  /* to forbid a solution means that
   * X0 != sol0 OR X1 != sol1 OR ... OR X81 != sol81
   * then each Xi != soli is transformed as in the function
   * add_negative.
   * So we finally have a big clause of 9*9*4 literals.
   * The code below does that.
   * it will add ii != v in each loop.
   * ii will be the 4 bits of variables i and v the solution of cell i.
   * (the code is clear, no?)
   */
  for (i=0; i<9*9; i++) {
    int v = sol[i];
    int ii = i*4+1;
    *cc++ = (v&1)? ii*2+1 : ii*2; ii++;
    *cc++ = (v&2)? ii*2+1 : ii*2; ii++;
    *cc++ = (v&4)? ii*2+1 : ii*2; ii++;
    *cc++ = (v&8)? ii*2+1 : ii*2; ii++;
  }

  SAT_AddClause(m, c, 9*9*4, 0);
}

int main(int n, char **vv)
{
  int i;
  int j;
  int ret;
  int nb_sol;
  int line;
  int column;
  int cube;
  SAT_Manager m = SAT_InitManager();

  SAT_SetNumVariables(m, 9*9*4);

  if (n != 2) {
    printf("usage: zsudoku <sudoku file> ('-' for stdin)\n");
    printf("  The sudoku file is just 9*9 numbers. 0 means unknown.\n");
    printf("  You may use any space you want between numbers.\n");
    printf("http://sed.free.fr/complex/sudoku.html\n");
    return 0;
  }

  if (vv[1][0] != '-' || vv[1][1] != 0) {
    stdin = fopen(vv[1], "r");
    if (!stdin) {
      perror(vv[1]);
      return 0;
    }
  }

  /* let's force the cases to have values between 1 and 9 inclusively */
  for (i=0; i<9*9; i++)
    force_value(m, i);

  /* AND all vars on a line must be != */
  /* wonderful for for for */
  for (line=0; line<9; line++)
    for (i=0; i<9; i++)
      for (j=i+1; j<9; j++)
        add_var_diff(m, i+line*9, j+line*9);

  /* AND all vars on a column must be != */
  for (column=0; column<9; column++)
    for (i=0; i<9; i++)
      for (j=i+1; j<9; j++)
        add_var_diff(m, i*9+column, j*9+column);

  /* AND all vars in a cube must be != */
  for (cube=0; cube<9; cube++)
    for (i=0; i<9; i++)
      for (j=i+1; j<9; j++) {
        /* don't add already added stuff (same line or same column) */
        if (i/3 == j/3) continue;
        if (i%3 == j%3) continue;
        do_cube(m, cube, i, j);
      }

  /* AND read a sudoku, which means to force the value of variables
   * for which the sudoku provides a non-null value
   */
  for (i=0; i<9*9; i++) {
    int dv = 0;
    fscanf(stdin, "%d", &dv);
    if (!dv) continue;
    add_positive(m, i, dv);
  }

  nb_sol = 0;

  /* solve it to death, until no more solution is found */
  do {
    printf("checking new solution\n");
    ret = sat_solve(m);
    if (ret) {
      /* if there was a solution, we forbid the solver to return it the
       * next time we run it
       */
      forbid_sol(m);
      SAT_Reset(m);
    }
    nb_sol += !!ret;
  } while (ret);

  SAT_ReleaseManager(m);

  printf("%d solutions\n", nb_sol);

  return 0;
}
